"""Macro template."""

from rimseval.processor import CRDFileProcessor


def calc(crd: CRDFileProcessor) -> None:
    """Macro template.

    :param crd: CRD file processor that will be passed to the macro.
    """
    # your code goes here
